/*
 * Copyright 2018 Outfox, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.outfoxx.swiftpoet.test

import io.outfoxx.swiftpoet.CodeWriter
import io.outfoxx.swiftpoet.DeclaredTypeName.Companion.typeName
import io.outfoxx.swiftpoet.INT
import io.outfoxx.swiftpoet.Modifier
import io.outfoxx.swiftpoet.SET
import io.outfoxx.swiftpoet.STRING
import io.outfoxx.swiftpoet.TypeAliasSpec
import io.outfoxx.swiftpoet.TypeName
import io.outfoxx.swiftpoet.TypeVariableName.Companion.bound
import io.outfoxx.swiftpoet.TypeVariableName.Companion.typeVariable
import io.outfoxx.swiftpoet.parameterizedBy
import io.outfoxx.swiftpoet.tag
import org.hamcrest.CoreMatchers.equalTo
import org.hamcrest.CoreMatchers.hasItems
import org.hamcrest.MatcherAssert.assertThat
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import java.io.StringWriter

@DisplayName("TypeAliasSpec Tests")
class TypeAliasSpecTests {

  @Test
  @DisplayName("Tags on builders can be retrieved on builders and built specs")
  fun testTags() {
    val testAliasBuilder = TypeAliasSpec.builder("Test", STRING)
      .tag(5)
    val testAlias = testAliasBuilder.build()

    assertThat(testAliasBuilder.tags[Integer::class] as? Int, equalTo(5))
    assertThat(testAlias.tag(), equalTo(5))
  }

  @Test
  @DisplayName("Generates documentation before type definition")
  fun testGenDoc() {
    val testAlias = TypeAliasSpec.builder("MyNumber", INT)
      .addDoc("this is a comment\n")
      .build()

    val out = StringWriter()
    testAlias.emit(CodeWriter(out))

    assertThat(
      out.toString(),
      equalTo(
        """
            /**
             * this is a comment
             */
            typealias MyNumber = Swift.Int

        """.trimIndent()
      )
    )
  }

  @Test
  @DisplayName("Generates attributes before type definition")
  fun testGenAttrs() {
    val testAlias = TypeAliasSpec.builder("MyNumber", INT)
      .addAttribute("available", "swift 5.1")
      .build()

    val out = StringWriter()
    testAlias.emit(CodeWriter(out))

    assertThat(
      out.toString(),
      equalTo(
        """
            @available(swift 5.1)
            typealias MyNumber = Swift.Int

        """.trimIndent()
      )
    )
  }

  @Test
  @DisplayName("Generates modifiers in order")
  fun testGenModifiersInOrder() {
    val testAlias = TypeAliasSpec.builder("MyNumber", INT)
      .addModifiers(Modifier.PUBLIC)
      .build()

    val out = StringWriter()
    testAlias.emit(CodeWriter(out))

    assertThat(
      out.toString(),
      equalTo(
        """
            public typealias MyNumber = Swift.Int

        """.trimIndent()
      )
    )
  }

  @Test
  @DisplayName("Generates simple alias")
  fun testSimpleAlias() {
    val testAlias = TypeAliasSpec.builder("MyNumber", INT)
      .build()

    val out = StringWriter()
    testAlias.emit(CodeWriter(out))

    assertThat(
      out.toString(),
      equalTo(
        """
            typealias MyNumber = Swift.Int

        """.trimIndent()
      )
    )
  }

  @Test
  @DisplayName("Generates generic alias")
  fun testGenericAlias() {
    val typeVar = typeVariable("A", bound(typeName(".Test")))
    val testAlias = TypeAliasSpec.builder("StringSet", SET.parameterizedBy(STRING, typeVar))
      .addTypeVariable(typeVar)
      .build()

    val out = StringWriter()
    testAlias.emit(CodeWriter(out))

    assertThat(
      out.toString(),
      equalTo(
        """
            typealias StringSet<A : Test> = Swift.Set<Swift.String, A>

        """.trimIndent()
      )
    )
  }

  @Test
  @DisplayName("toBuilder copies all fields")
  fun testToBuilder() {
    val testAliasBldr = TypeAliasSpec.builder("Test", INT)
      .addDoc("this is a comment\n")
      .addModifiers(Modifier.PUBLIC)
      .addTypeVariable(typeVariable("A", bound(typeName(".Test"))))
      .build()
      .toBuilder()

    assertThat(testAliasBldr.name, equalTo("Test"))
    assertThat(testAliasBldr.type, equalTo<TypeName>(INT))
    assertThat(testAliasBldr.doc.formatParts, hasItems("this is a comment\n"))
    assertThat(testAliasBldr.modifiers, hasItems(Modifier.PUBLIC))
    assertThat(testAliasBldr.typeVariables.map { it.name }, hasItems("A"))
  }
}
